"""Pytest fixtures and settings used in the qnexus integration tests."""

from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Generator, cast

import pytest
from hugr.package import Package
from pytket.circuit import Circuit
from pytket.qir import pytket_to_qir  # type: ignore[attr-defined]
from pytket.wasm.wasm import WasmFileHandler
from quantinuum_schemas.models.backend_config import BackendConfig

import qnexus as qnx
from qnexus.config import CONFIG
from qnexus.models.references import CircuitRef


@contextmanager
def make_authenticated_nexus(
    user_email: str = CONFIG.qa_user_email,
    user_password: str = CONFIG.qa_user_password,
) -> Generator[None, None, None]:
    """Authenticate the qnexus client."""
    try:
        qnx.auth._request_tokens(user_email, user_password)
        yield
    finally:
        qnx.auth.logout()


@pytest.fixture(scope="session")
def _authenticated_nexus(
    qa_project_name: str,
    qa_circuit_name: str,
    qa_circuit_name_2: str,
    qa_team_name: str,
    qa_compile_job_name: str,
    qa_execute_job_name: str,
    qa_wasm_module_name: str,
    qa_hugr_name: str,
    qa_qir_name: str,
    circuit: Circuit,
) -> Generator[None, None, None]:
    """Authenticated nexus instance fixture."""

    with make_authenticated_nexus():
        test_desc = f"This can be safely deleted. Test Run: {datetime.now()}"

        my_proj = qnx.projects.create(name=qa_project_name, description=test_desc)

        qnx.projects.add_property(
            name="QA_test_prop",
            property_type="string",
            project=my_proj,
        )

        qnx.teams.create(name=qa_team_name, description=test_desc)

        my_new_circuit = qnx.circuits.upload(
            circuit=Circuit(2, 2).H(0).CX(0, 1).measure_all(),
            name=qa_circuit_name,
            description=test_desc,
            project=my_proj,
        )

        my_other_circuit = qnx.circuits.upload(
            # re-uploading with a unique name to avoid conflicts
            circuit=Circuit(2, 2).H(0).CX(0, 1).measure_all(),
            name=qa_circuit_name_2,
            description=test_desc,
            project=my_proj,
        )

        compile_job_ref = qnx.start_compile_job(
            programs=[my_other_circuit],
            name=qa_compile_job_name,
            description=test_desc,
            project=my_proj,
            backend_config=qnx.AerConfig(),
            skip_intermediate_circuits=False,
        )

        execute_job_ref = qnx.start_execute_job(
            programs=[my_new_circuit],
            name=qa_execute_job_name,
            description=test_desc,
            project=my_proj,
            backend_config=qnx.AerConfig(),
            n_shots=[10],
        )

        wasm_path = Path("examples/basics/data/add_one.wasm").resolve()
        wfh = WasmFileHandler(filepath=str(wasm_path))

        qnx.wasm_modules.upload(
            wasm_module_handler=wfh,
            name=qa_wasm_module_name,
            project=my_proj,
        )

        hugr_path = Path("integration/data/hugr_example.hugr").resolve()
        hugr_package = Package.from_bytes(hugr_path.read_bytes())

        qnx.hugr.upload(
            hugr_package=hugr_package,
            name=qa_hugr_name,
            project=my_proj,
        )

        qnx.jobs.wait_for(compile_job_ref)
        qnx.jobs.wait_for(execute_job_ref)

        qir_bitcode = pytket_to_qir(circuit, name=qa_qir_name)
        assert isinstance(qir_bitcode, bytes)
        qnx.qir.upload(qir=qir_bitcode, name=qa_qir_name, project=my_proj)

        yield

        qnx.projects.update(my_proj, archive=True)
        qnx.projects.delete(my_proj)


@pytest.fixture(scope="session")
def _authenticated_nexus_circuit_ref(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> Generator[CircuitRef, None, None]:
    """Starting with authenticated nexus instance, yield a CircuitRef
    for use in tests."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    my_new_circuit = qnx.circuits.upload(
        circuit=Circuit(2, 2).H(0).CX(0, 1).measure_all(),
        name=f"qnexus_integration_additional_test_circuit_{datetime.now()}",
        description=f"This can be safely deleted. Test Run: {datetime.now()}",
        project=my_proj,
    )

    yield my_new_circuit


@pytest.fixture(scope="session", name="qa_project_name")
def qa_project_name_fixture() -> str:
    """A name for uniquely identifying a project owned by the Nexus QA user."""
    return f"qnexus_integration_test_project_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_team_name")
def qa_team_name_fixture() -> str:
    """A name for uniquely identifying a team for the Nexus QA user."""
    return f"qnexus_integration_test_team_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_circuit_name")
def qa_circuit_name_fixture() -> str:
    """A name for uniquely identifying a circuit owned by the Nexus QA user,
    in the project specified by qa_project_name."""
    return f"qnexus_integration_test_circuit_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_circuit_name_2")
def qa_circuit_name2_fixture() -> str:
    """A name for uniquely identifying a circuit owned by the Nexus QA user,
    in the project specified by qa_project_name."""
    return f"qnexus_integration_test_circuit2_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_compile_job_name")
def qa_compile_job_name_fixture() -> str:
    """A name for uniquely identifying a compile job owned by the Nexus QA user,
    in the project specified by qa_project_name."""
    return f"qnexus_integration_test_compile_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_execute_job_name")
def qa_execute_job_name_fixture() -> str:
    """A name for uniquely identifying an execute job owned by the Nexus QA user,
    in the project specified by qa_project_name."""
    return f"qnexus_integration_test_execute_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_wasm_module_name")
def qa_wasm_module_name_fixture() -> str:
    """A name for uniquely identifying a WASM module owned by the Nexus QA user."""
    return f"qnexus_integration_test_wasm_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_hugr_name")
def qa_hugr_name_fixture() -> str:
    """A name for uniquely identifying a HUGR owned by the Nexus QA user."""
    return f"qnexus_integration_test_hugr_{datetime.now()}"


@pytest.fixture(scope="session", name="qa_qir_name")
def qa_qir_name_fixture() -> str:
    """A name for uniquely identifying a QIR owned by the Nexus QA user."""
    return f"qnexus_integration_test_qir_{datetime.now()}"


def get_backend_config_name(backend_config: qnx.BackendConfig) -> str:
    name = backend_config.__class__.__name__
    if hasattr(backend_config, "backend_name"):
        name += f"({backend_config.backend_name})"
    if hasattr(backend_config, "device_name"):
        name += f"({backend_config.device_name})"
    return name


@pytest.fixture(
    scope="session",
    params=[
        # Nexus-hosted
        qnx.AerConfig(),
        qnx.AerStateConfig(),
        qnx.AerUnitaryConfig(),
        qnx.BraketConfig(local=True),
        qnx.QuantinuumConfig(device_name="H1-1LE"),
        qnx.ProjectQConfig(),
        qnx.QulacsConfig(),
        # Non Nexus-hosted
        qnx.IBMQConfig(
            backend_name="ibm_sherbrooke",
            instance=(
                "crn:v1:bluemix:public:quantum-computing:us-east:"
                "a/18f63f4565ef4a40851959792418cbf2:"
                "37bf946a-6349-47df-a092-3fbd5b92dbf2::"
            ),
        ),
        qnx.IBMQEmulatorConfig(
            backend_name="ibm_sherbrooke",
            instance=(
                "crn:v1:bluemix:public:quantum-computing:us-east:"
                "a/18f63f4565ef4a40851959792418cbf2:"
                "37bf946a-6349-47df-a092-3fbd5b92dbf2::"
            ),
        ),
        qnx.QuantinuumConfig(device_name="H1-1SC"),  # Cluster-hosted
    ],
    ids=get_backend_config_name,
)
def backend_config(request: pytest.FixtureRequest) -> BackendConfig:
    """Fixture to provide an instance of all BackendConfigs for testing."""
    return cast(BackendConfig, request.param)


@pytest.fixture(scope="session")
def circuit() -> Circuit:
    """A pytket circuit"""

    circuit = Circuit(3)
    circuit.H(0)
    for i, j in zip(circuit.qubits[:-1], circuit.qubits[1:]):
        circuit.CX(i, j)
    circuit.measure_all()
    return circuit



"""Client API for Nexus."""

import typing
from logging import getLogger
import httpx

from qnexus.client.utils import read_token, write_token
from qnexus.config import CONFIG
from qnexus.exceptions import AuthenticationError

logger = getLogger()


class AuthHandler(httpx.Auth):
    """Custom nexus auth handler"""

    cookies: httpx.Cookies

    def __init__(self) -> None:
        self.cookies = httpx.Cookies()
        self.reload_tokens()

        super().__init__()

    def reload_tokens(self) -> None:
        """Clear tokens and attempt to reload from the file system."""
        try:
            self.cookies.clear()
            token = read_token("refresh_token")
            self.cookies.set("myqos_oat", token, domain=CONFIG.domain)
            id_token = read_token("access_token")
            self.cookies.set("myqos_id", id_token, domain=CONFIG.domain)
        except FileNotFoundError:
            pass

    def auth_flow(
        self, request: httpx.Request
    ) -> typing.Generator[httpx.Request, httpx.Response, None]:
        self.cookies.set_cookie_header(request)

        response = yield request

        _print_sunset_header_warning(request, response)

        if response.status_code == 401:
            if self.cookies.get("myqos_oat") is None:
                try:
                    token = read_token(
                        "refresh_token",
                    )
                    self.cookies.set("myqos_oat", token, domain=CONFIG.domain)
                except FileNotFoundError as exc:
                    raise AuthenticationError(
                        "Not authenticated. Please run `qnx login` in your terminal."
                    ) from exc

            auth_response = yield self.build_refresh_request()
            if auth_response.status_code == 401:
                raise AuthenticationError(
                    "Not authenticated. Please run `qnx login` in your terminal."
                )

            auth_response.raise_for_status()
            self.cookies.extract_cookies(auth_response)

            write_token(
                "access_token",
                self.cookies.get("myqos_id", domain=CONFIG.domain) or "",
            )
            if request.headers.get("cookie"):
                request.headers.pop("cookie")
            self.cookies.set_cookie_header(request)
            yield request

    def build_refresh_request(self) -> httpx.Request:
        """Build the request for refreshing the id token."""
        self.cookies.delete("myqos_id")  # We need to delete any existing id token first
        return httpx.Request(
            method="POST",
            url=f"{CONFIG.url}/auth/tokens/refresh",
            cookies=self.cookies,
        )


_nexus_client: httpx.Client | None = None


def get_nexus_client(reload: bool = False) -> httpx.Client:
    """Getter function for the nexus client.
    Args:
        reload: If True, reconstruct the client (and auth tokens).
    Returns:
        httpx.Client: The nexus client.
    """
    global _nexus_client
    if _nexus_client is None or reload:
        _auth_handler = AuthHandler()
        _auth_handler.reload_tokens()

        _nexus_client = httpx.Client(
            base_url=CONFIG.url,
            auth=_auth_handler,
            timeout=None,
            verify=CONFIG.httpx_verify,
        )
    return _nexus_client


def _print_sunset_header_warning(
    request: httpx.Request, response: httpx.Response
) -> None:
    sunset_header = response.headers.get("sunset")
    if sunset_header:
        logger.warning(
            f"Your are currently using a deprecated API endpoint ({request.url}) that will be deleted on {sunset_header}. After this date your current qnexus version may stop functioning. Please update to a later qnexus version to resolve the issue."
        )



"""Client API for teams in Nexus."""

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.models.references import DataframableList, TeamRef


def get_all() -> DataframableList[TeamRef]:
    """No fuzzy name matching."""
    res = get_nexus_client().get(
        "/api/v5/user/teams",
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    return DataframableList(
        [
            TeamRef(
                id=team["id"],
                name=team["team_name"],
                description=team["description"],
            )
            for team in res.json()
        ]
    )


def get(name: str) -> TeamRef:
    """
    Get a single team using filters. Throws an exception if the filters do not
    match exactly one object.
    """
    res = get_nexus_client().get("/api/v5/user/teams", params={"name": name})

    if res.status_code == 404 or res.json() == []:
        raise qnx_exc.ZeroMatches

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    teams_list = [
        TeamRef(
            id=team["id"],
            name=team["team_name"],
            description=team["description"],
        )
        for team in res.json()
    ]

    if len(teams_list) > 1:
        raise qnx_exc.NoUniqueMatch

    return teams_list[0]


def _fetch_by_id(team_id: str) -> TeamRef:
    """
    Get a single team by id.
    """
    res = get_nexus_client().get(f"/api/v5/user/teams/{team_id}")

    if res.status_code == 404:
        raise qnx_exc.ZeroMatches

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    team_dict = res.json()

    return TeamRef(
        id=team_dict["id"],
        name=team_dict["team_name"],
        description=team_dict["description"],
    )


def create(name: str, description: str | None = None) -> TeamRef:
    """Create a team in Nexus."""

    resp = get_nexus_client().post(
        "api/v5/user/teams/new",
        json={
            "team_name": name,
            "description": description,
        },
    )

    if resp.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    team_dict = resp.json()
    return TeamRef(
        id=team_dict["id"],
        name=team_dict["team_name"],
        description=team_dict["description"],
    )



"""Status types for Nexus Jobs."""

from datetime import datetime
from typing import Any, Dict, NamedTuple, Optional

import pandas as pd
from pytket.backends.status import (
    StatusEnum,
)


class JobStatus(NamedTuple):
    """The status of a job along with an optional description.

    Optionally can also include extra fields such as:
    * Detailed error information.
    * Timestamps for changes in status.
    * Queue position.
    """

    status: StatusEnum
    message: str = ""
    error_detail: Optional[str] = None

    # Timestamp for when a status was last entered.
    completed_time: Optional[datetime] = None
    queued_time: Optional[datetime] = None
    submitted_time: Optional[datetime] = None
    running_time: Optional[datetime] = None
    cancelled_time: Optional[datetime] = None
    error_time: Optional[datetime] = None

    queue_position: Optional[int] = None

    @classmethod
    def from_dict(cls, dic: Dict[str, Any]) -> "JobStatus":
        """Construct from JSON serializable dictionary."""
        invalid = ValueError(f"Dictionary invalid format for JobStatus: {dic}")
        if "message" not in dic or "status" not in dic:
            raise invalid

        try:
            status = next(s for s in StatusEnum if dic["status"] == s.name)
        except StopIteration as err:
            raise invalid from err

        error_detail = dic.get("error_detail", None)

        def read_optional_datetime(key: str) -> datetime | None:
            x = dic.get(key)
            return datetime.fromisoformat(x) if x is not None else None

        completed_time = read_optional_datetime("completed_time")
        queued_time = read_optional_datetime("queued_time")
        submitted_time = read_optional_datetime("submitted_time")
        running_time = read_optional_datetime("running_time")
        cancelled_time = read_optional_datetime("cancelled_time")
        error_time = read_optional_datetime("error_time")

        queue_position = dic.get("queue_position", None)

        return cls(
            status,
            dic["message"],
            error_detail,
            completed_time,
            queued_time,
            submitted_time,
            running_time,
            cancelled_time,
            error_time,
            queue_position,
        )

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame.from_dict(
            self._asdict(),
            orient="index",
        ).T



"""CLI for qnexus."""

import click

from ..client import auth as _auth


@click.command()
def login() -> None:
    """Log in to quantinuum nexus using your web browser."""
    click.echo(_auth.login())  # type: ignore[func-returns-value]


@click.command()
def logout() -> None:
    """Log out of quantinuum nexus."""
    click.echo(_auth.logout())  # type: ignore[func-returns-value]



"""The qnexus package."""

import warnings

import nest_asyncio  # type: ignore
from quantinuum_schemas.models.backend_config import (
    AerConfig,
    AerStateConfig,
    AerUnitaryConfig,
    BackendConfig,
    BraketConfig,
    IBMQConfig,
    IBMQEmulatorConfig,
    ProjectQConfig,
    QuantinuumConfig,
    QulacsConfig,
    SeleneClassicalReplayConfig,
    SeleneCoinflipConfig,
    SeleneLeanConfig,
    SeleneQuestConfig,
    SeleneStimConfig,
)

import qnexus.models as models
from qnexus import context, filesystem
from qnexus.client import (
    auth,
    circuits,
    credentials,
    devices,
    hugr,
    jobs,
    projects,
    qir,
    quotas,
    roles,
    teams,
    users,
    wasm_modules,
)
from qnexus.client.auth import login, login_with_credentials, logout
from qnexus.client.jobs import compile, execute
from qnexus.client.jobs._compile import start_compile_job
from qnexus.client.jobs._execute import start_execute_job

warnings.filterwarnings("default", category=DeprecationWarning, module=__name__)

# This is necessary for use in Jupyter notebooks to allow for nested asyncio loops
try:
    nest_asyncio.apply()
except (RuntimeError, ValueError):
    # May fail in some cloud environments: ignore.
    pass

__all__ = [
    "context",
    "roles",
    "auth",
    "circuits",
    "credentials",
    "devices",
    "hugr",
    "qir",
    "jobs",
    "projects",
    "quotas",
    "teams",
    "compile",
    "execute",
    "users",
    "wasm_modules",
    "filesystem",
    "start_compile_job",
    "start_execute_job",
    "login",
    "login_with_credentials",
    "logout",
    "AerConfig",
    "AerStateConfig",
    "AerUnitaryConfig",
    "BackendConfig",
    "BraketConfig",
    "IBMQConfig",
    "IBMQEmulatorConfig",
    "ProjectQConfig",
    "QuantinuumConfig",
    "QulacsConfig",
    "SeleneCoinflipConfig",
    "SeleneClassicalReplayConfig",
    "SeleneLeanConfig",
    "SeleneStimConfig",
    "SeleneQuestConfig",
    "models",
]



"""Test basic functionality relating to the qir module."""

from collections import Counter
from datetime import datetime

from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Bit, Circuit
from pytket.qir import pytket_to_qir  # type: ignore[attr-defined]

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import QIRRef


def test_qir_create_and_update(
    _authenticated_nexus: None,
    qa_project_name: str,
    circuit: Circuit,
) -> None:
    """Test that we can create a qir and add a property value."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    qir_name = f"QA_test_qir_{datetime.now()}"

    qir_bitcode = pytket_to_qir(circuit, name=qir_name)
    assert isinstance(qir_bitcode, bytes)
    my_new_qir = qnx.qir.upload(qir=qir_bitcode, name=qir_name, project=my_proj)

    assert isinstance(my_new_qir, QIRRef)

    test_property_name = "QA_test_prop"
    test_prop_value = "foo"

    updated_qir_ref = qnx.qir.update(
        ref=my_new_qir,
        properties=PropertiesDict({test_property_name: test_prop_value}),
    )

    assert updated_qir_ref.annotations.properties[test_property_name] == test_prop_value


def test_qir_download(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_qir_name: str,
) -> None:
    """Test that QIR bytes can be downloaded from an uploaded QIR module."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_qir_ref = qnx.qir.get(name_like=qa_qir_name, project=my_proj)

    qir_bytes = my_qir_ref.download_qir()
    assert isinstance(qir_bytes, bytes)


def test_qir_get_by_id(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_qir_name: str,
) -> None:
    """Test that we can get a QIRRef by its ID."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_qir_ref = qnx.qir.get(name_like=qa_qir_name, project=my_proj)

    qir_ref_by_id = qnx.qir.get(id=my_qir_ref.id)

    assert qir_ref_by_id == my_qir_ref


def test_qir_get_all(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_qir_name: str,
) -> None:
    """Test that we can get all qirRefs in a project."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    qirs = qnx.qir.get_all(project=my_proj)

    assert qirs.count() >= 1
    assert isinstance(qirs.list()[0], QIRRef)


def test_execution(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_qir_name: str,
) -> None:
    """Test the execution of a QIR program."""

    device_name = "H1-1SC"  # Syntax checker - no results

    project_ref = qnx.projects.get_or_create(name=qa_project_name)

    qir_program_ref = qnx.qir.get(name_like=qa_qir_name)

    job_ref = qnx.start_execute_job(
        programs=[qir_program_ref],
        n_shots=[10],
        backend_config=qnx.QuantinuumConfig(device_name=device_name),
        project=project_ref,
        name=f"QA Test QIR job from {datetime.now()}",
    )

    qnx.jobs.wait_for(job_ref)

    results = qnx.jobs.results(job_ref)

    assert len(results) == 1
    result_ref = results[0]

    assert isinstance(result_ref.download_backend_info(), BackendInfo)
    assert isinstance(result_ref.get_input(), QIRRef)

    assert result_ref.get_input().id == qir_program_ref.id

    qir_result = qnx.jobs.results(job_ref)[0].download_result()
    assert isinstance(qir_result, BackendResult)
    assert qir_result.get_counts() == Counter({(0, 0, 0): 10})
    assert qir_result.get_bitlist() == [Bit("c", 2), Bit("c", 1), Bit("c", 0)]



"""Test basic functionality relating to the wasm_modules module."""

from datetime import datetime
from pathlib import Path

import pandas as pd
from pytket.backends.status import StatusEnum
from pytket.circuit import Circuit
from pytket.wasm.wasm import WasmFileHandler

import qnexus as qnx
from qnexus.models.references import WasmModuleRef


def test_wasm_download(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that valid WASM can be extracted from an uploaded WASM module."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    wasm_path = Path("examples/basics/data/add_one.wasm").resolve()
    wfh = WasmFileHandler(filepath=str(wasm_path))
    qa_wasm_module_name_fixture = f"qnexus_integration_test_wasm_{datetime.now()}"

    qnx.wasm_modules.upload(
        wasm_module_handler=wfh,
        name=qa_wasm_module_name_fixture,
        project=my_proj,
    )

    wasm_ref = qnx.wasm_modules.get(name_like=qa_wasm_module_name_fixture)
    assert isinstance(wasm_ref, WasmModuleRef)
    downloaded_wasm_module_handler = wasm_ref.download_wasm_contents()

    downloaded_wasm_module_handler.check()
    assert downloaded_wasm_module_handler.functions == wfh.functions


def test_wasm_flow(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test the flow for executing a simple WASM circuit on H1-1LE."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    wasm_path = Path("examples/basics/data/add_one.wasm").resolve()
    wfh = WasmFileHandler(filepath=str(wasm_path))
    qa_wasm_module_name_fixture = f"qnexus_integration_test_wasm_{datetime.now()}"

    qnx.wasm_modules.upload(
        wasm_module_handler=wfh,
        name=qa_wasm_module_name_fixture,
        project=my_proj,
    )

    my_wasm_db_matches = qnx.wasm_modules.get_all()
    assert isinstance(my_wasm_db_matches.summarize(), pd.DataFrame)
    assert isinstance(next(my_wasm_db_matches), WasmModuleRef)

    wasm_ref = qnx.wasm_modules.get(name_like=qa_wasm_module_name_fixture)
    assert isinstance(wasm_ref, WasmModuleRef)

    wasm_ref_2 = qnx.wasm_modules.get(id=wasm_ref.id)
    assert wasm_ref == wasm_ref_2

    circuit = Circuit(1)
    a = circuit.add_c_register("a", 8)
    circuit.add_wasm_to_reg("add_one", wfh, [a], [a])
    circuit.measure_all()
    qa_wasm_circuit_name_fixture = (
        f"qnexus_integration_test_wasm_circuit_{datetime.now()}"
    )

    wasm_circuit_ref = qnx.circuits.upload(
        name=qa_wasm_circuit_name_fixture,
        circuit=circuit,
        project=my_proj,
    )

    execute_job_ref = qnx.start_execute_job(
        programs=[wasm_circuit_ref],
        name=f"qnexus_integration_test_wasm_execute_job_{datetime.now()}",
        n_shots=[100],
        backend_config=qnx.QuantinuumConfig(
            device_name="H1-Emulator",
        ),
        wasm_module=wasm_ref,
        project=my_proj,
    )

    qnx.jobs.wait_for(execute_job_ref)

    assert qnx.jobs.status(execute_job_ref).status == StatusEnum.COMPLETED



"""Client API for projects in Nexus."""

import random
from datetime import datetime
from typing import Any, Literal, Union, cast
from uuid import UUID

import pandas as pd
from pytket.backends.status import WAITING_STATUS, StatusEnum

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import get_active_project
from qnexus.models import Property
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (  # PropertiesFilter, # Not yet implemented
    ArchivedFilter,
    CreatorFilter,
    FuzzyNameFilter,
    PaginationFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import DataframableList, ProjectRef

# Colour-blind friendly colours from https://www.nature.com/articles/nmeth.1618
_COLOURS = ["#e69f00", "#56b4e9", "#009e73", "#f0e442", "#0072b2", "#d55e00", "#cc79a7"]


class Params(
    ScopeFilter,
    SortFilter,
    PaginationFilter,
    FuzzyNameFilter,
    CreatorFilter,
    # PropertiesFilter, # Not yet implemented
    TimeFilter,
    ArchivedFilter,
):
    """Params for filtering projects"""


def get_all(
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    is_archived: bool = False,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[ProjectRef]:
    """Get a NexusIterator over projects with optional filters."""

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        is_archived=is_archived,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(
        by_alias=True,
        exclude_unset=True,
        exclude_none=True,
    )

    return NexusIterator(
        resource_type="Project",
        nexus_url="/api/projects/v1beta",
        params=params,
        wrapper_method=_to_projectref,
        nexus_client=get_nexus_client(),
    )


def _to_projectref(data: dict[str, Any]) -> DataframableList[ProjectRef]:
    return DataframableList(
        [
            ProjectRef(
                id=project["id"],
                annotations=Annotations.from_dict(project["attributes"]),
                contents_modified=project["attributes"]["contents_modified"],
                archived=project["attributes"]["archived"],
            )
            for project in data["data"]
        ]
    )


def get(
    *,
    id: Union[str, UUID, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    is_archived: bool = False,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> ProjectRef:
    """
    Get a single project using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        is_archived=is_archived,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


def get_or_create(
    name: str,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> ProjectRef:
    """Get a project reference if the projects exists (by name),
    otherwise create a new project using the supplied description and properties."""
    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    )
    try:
        return get(name_like=annotations.name)
    except qnx_exc.ZeroMatches:
        return create(
            name=annotations.name,
            description=annotations.description,
            properties=annotations.properties,
        )


def _fetch_by_id(project_id: UUID | str, scope: ScopeFilterEnum | None) -> ProjectRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(scope=scope).model_dump(
        by_alias=True, exclude_unset=True, exclude_none=True
    )

    res = get_nexus_client().get(f"/api/projects/v1beta/{project_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    return ProjectRef(
        id=res_dict["data"]["id"],
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        contents_modified=res_dict["data"]["attributes"]["contents_modified"],
        archived=res_dict["data"]["attributes"]["archived"],
    )


def create(
    name: str,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> ProjectRef:
    """Create a new project in Nexus."""
    attributes = {}
    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": {},
            "type": "project",
        }
    }

    res = get_nexus_client().post("/api/projects/v1beta", json=req_dict)

    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return ProjectRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        contents_modified=res_data_dict["attributes"]["contents_modified"],
        archived=res_data_dict["attributes"]["archived"],
    )


def add_property(
    name: str,
    property_type: Literal["bool", "int", "float", "string"],
    project: ProjectRef | None = None,
    description: str | None = None,
    required: bool = False,
) -> None:
    """Add a property definition to a project."""
    project = project or get_active_project(project_required=True)
    assert project, "ProjectRef required."

    # For now required to add properties in a seperate API step
    props_req_dict = {
        "data": {
            "attributes": {
                "name": name,
                "description": description,
                "property_type": property_type,
                "required": required,
                "color": random.choice(_COLOURS),
            },
            "relationships": {
                "project": {"data": {"id": str(project.id), "type": "project"}}
            },
            "type": "property",
        }
    }
    props_res = get_nexus_client().post(
        "/api/property_definitions/v1beta", json=props_req_dict
    )

    if props_res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=props_res.text, status_code=props_res.status_code
        )


def get_properties(project: ProjectRef | None = None) -> DataframableList[Property]:
    """Get the property definitions for the Project."""
    # This will get all the properties (even if there are more than the page size),
    # but could be refactored if needed.

    project = project or get_active_project(project_required=True)
    assert project, "ProjectRef required."

    all_project_properties_iterator = NexusIterator(
        resource_type="Property",
        nexus_url="/api/property_definitions/v1beta",
        params={"filter[project][id]": str(project.id)},
        wrapper_method=_to_property,
        nexus_client=get_nexus_client(),
    )

    return all_project_properties_iterator.list()


def _to_property(data: dict[str, Any]) -> DataframableList[Property]:
    return DataframableList(
        [
            Property(
                annotations=Annotations.from_dict(property_dict["attributes"]),
                property_type=property_dict["attributes"]["property_type"],
                required=property_dict["attributes"]["required"],
                color=property_dict["attributes"]["color"],
                id=property_dict["id"],
            )
            for property_dict in data["data"]
        ]
    )


def summarize(project: ProjectRef | None = None) -> pd.DataFrame:
    """Summarize the current state of a project."""
    import qnexus.client.jobs as jobs_client

    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    all_jobs = jobs_client.get_all(project=project).list()

    return pd.DataFrame(
        {
            "project": project.annotations.name,
            "total_jobs": len(all_jobs),
            "pending_jobs": len(
                [job for job in all_jobs if job.last_status in WAITING_STATUS]
            ),
            "cancelled_jobs": len(
                [job for job in all_jobs if job.last_status == StatusEnum.CANCELLED]
            ),
            "errored_jobs": len(
                [job for job in all_jobs if job.last_status == StatusEnum.ERROR]
            ),
            "completed_jobs": len(
                [job for job in all_jobs if job.last_status == StatusEnum.COMPLETED]
            ),
        },
        index=[0],
    )


def update(
    project: ProjectRef,
    name: str | None = None,
    description: str | None = None,
    archive: bool = False,
) -> ProjectRef:
    """Update the details of a project."""
    req_dict = {
        "data": {
            "attributes": {
                "name": name,
                "description": description,
                "archived": archive,
            },
            "type": "project",
            "relationships": {},
        }
    }

    res = get_nexus_client().patch(f"/api/projects/v1beta/{project.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return ProjectRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        contents_modified=res_data_dict["attributes"]["contents_modified"],
        archived=res_data_dict["attributes"]["archived"],
    )


def delete(project: ProjectRef) -> None:
    """Delete a project and all associated data in Nexus.
    Project must be archived first.
    WARNING: this will delete all data associated with the project.
    """
    res = get_nexus_client().delete(
        url=f"/api/projects/v1beta/{project.id}", params={"scope": "user"}
    )

    if res.status_code != 204:
        raise qnx_exc.ResourceDeleteFailed(
            message=res.text, status_code=res.status_code
        )



"""Test the context management system."""

from collections import OrderedDict
from datetime import datetime
from uuid import uuid4

import pytest

from qnexus.context import (
    deactivate_project,
    get_active_project,
    get_active_properties,
    merge_project_from_context,
    merge_properties_from_context,
    set_active_project_token,
    using_project,
    using_properties,
)
from qnexus.models.annotations import Annotations, PropertiesDict
from qnexus.models.references import ProjectRef


def test_attach_project() -> None:
    """Test that we can set a Project in the global context."""
    project_id = uuid4()
    project = ProjectRef(
        id=project_id,
        annotations=Annotations(name=""),
        contents_modified=datetime.now(),
        archived=False,
    )

    token = set_active_project_token(project=project)

    ctx_project = get_active_project()

    assert project == ctx_project

    deactivate_project(token)


def test_attach_project_context_manager() -> None:
    """Test that we can set a Project via a context manager."""

    project_id = uuid4()
    project = ProjectRef(
        id=project_id,
        annotations=Annotations(name=""),
        contents_modified=datetime.now(),
        archived=False,
    )
    with using_project(project=project):
        ctx_project = get_active_project()

        assert project == ctx_project

    ctx_project = get_active_project()

    assert ctx_project is None


def test_attach_properties() -> None:
    """Test that we can set properties in the global context."""
    base_properties = {"foo": 3}
    with using_properties(**base_properties):
        more_properties = {"bar": 5}
        with using_properties(**more_properties):
            ctx_properties = get_active_properties()

            combined_properties = {}
            combined_properties.update(base_properties)
            combined_properties.update(more_properties)

            assert combined_properties == ctx_properties

        ctx_properties = get_active_properties()
        assert base_properties == ctx_properties

    ctx_properties = get_active_properties()

    assert ctx_properties == {}


def test_merge_project_from_context() -> None:
    """Test the decorator for merging a projectref from context or function arguments."""

    project = ProjectRef(
        id=uuid4(),
        annotations=Annotations(
            name="test_project",
        ),
        contents_modified=datetime.now(),
        archived=False,
    )

    @merge_project_from_context
    def func_wants_project(project: ProjectRef | None = None) -> ProjectRef:
        """Dummy function for testing the merge_project decorator"""
        assert project is not None
        assert isinstance(project, ProjectRef)
        return project

    with pytest.raises(AssertionError):
        func_wants_project()

    returned_project = func_wants_project(project=project)
    assert returned_project == project

    with using_project(project=project):
        returned_project = func_wants_project()
        assert returned_project == project

    other_project = ProjectRef(
        id=uuid4(),
        annotations=Annotations(
            name="test_project_2",
        ),
        contents_modified=datetime.now(),
        archived=False,
    )

    with using_project(project=project):
        # kwarg takes precedence
        returned_project = func_wants_project(project=other_project)
        assert returned_project == other_project


def test_merge_properties_from_context() -> None:
    """Test the decorator for merging properties from context or function arguments."""

    @merge_properties_from_context
    def func_wants_properties(
        properties: PropertiesDict | None = None,
    ) -> PropertiesDict:
        """Dummy function for testing the merge_properties decorator"""
        # Property decorator provides empty properties dict by default
        assert isinstance(properties, OrderedDict)
        return properties

    assert func_wants_properties() == PropertiesDict({})

    first_props = PropertiesDict({"1": 2, "shared_key": "world"})

    returned_props = func_wants_properties(properties=first_props)
    assert returned_props == first_props

    with using_properties(**first_props):
        returned_props = func_wants_properties()
        assert returned_props == first_props

    second_props = PropertiesDict({"bingo": 2, "shared_key": 2.01})

    with using_properties(**first_props):
        # kwarg props take precedence, but otherwise union is taken
        returned_props = func_wants_properties(properties=second_props)

        assert returned_props["1"] == first_props["1"]
        assert returned_props["bingo"] == second_props["bingo"]
        assert returned_props["shared_key"] == second_props["shared_key"]
        assert len(returned_props.keys()) == 3


def test_merge_only_properties_from_context() -> None:
    """Test the decorator for merging properties from context when no properties are
    passed as an argument."""

    @merge_properties_from_context
    def func_wants_properties(
        properties: PropertiesDict | None = None,
    ) -> PropertiesDict:
        """Dummy function for testing the merge_properties decorator"""
        # Property decorator provides empty properties dict by default
        assert isinstance(properties, OrderedDict)
        return properties

    assert func_wants_properties() == PropertiesDict({})

    test_props = PropertiesDict({"1": 2, "shared_key": "world"})

    with using_properties(**test_props):
        returned_props = func_wants_properties(properties=None)
        assert returned_props == test_props



"""Models for use by the client."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from logging import getLogger
from typing import Literal, Optional, Union
from uuid import UUID

import pandas as pd
from pydantic import BaseModel, ConfigDict, field_validator
from pytket.architecture import Architecture, FullyConnected
from pytket.backends.backendinfo import BackendInfo
from pytket.circuit import Node, OpType
from quantinuum_schemas.models.backend_config import (
    AerConfig,
    AerStateConfig,
    AerUnitaryConfig,
    BackendConfig,
    BraketConfig,
    IBMQConfig,
    IBMQEmulatorConfig,
    ProjectQConfig,
    QuantinuumConfig,
    QulacsConfig,
    SeleneClassicalReplayConfig,
    SeleneCoinflipConfig,
    SeleneLeanConfig,
    SeleneQuestConfig,
    SeleneStimConfig,
)
from quantinuum_schemas.models.backend_info import Register, StoredBackendInfo
from quantinuum_schemas.models.selene_config import (
    DepolarizingErrorModel,
    HeliosRuntime,
    NoErrorModel,
    SimpleRuntime,
)

from qnexus.models.annotations import Annotations
from qnexus.models.references import TeamRef, UserRef
from qnexus.models.utils import assert_never

logger = getLogger(__name__)

__all__ = [
    "AerConfig",
    "AerStateConfig",
    "AerUnitaryConfig",
    "BackendConfig",
    "BraketConfig",
    "IBMQConfig",
    "IBMQEmulatorConfig",
    "ProjectQConfig",
    "QuantinuumConfig",
    "QulacsConfig",
    "StoredBackendInfo",
    "SeleneCoinflipConfig",
    "SeleneClassicalReplayConfig",
    "SeleneLeanConfig",
    "SeleneStimConfig",
    "SeleneQuestConfig",
    "HeliosRuntime",
    "SimpleRuntime",
    "NoErrorModel",
    "DepolarizingErrorModel",
]


class CredentialIssuer(str, Enum):
    """An Issuer of credentials."""

    QUANTINUUM = "Quantinuum"
    QISKIT = "Qiskit"
    BRAKET = "Braket"


class Credential(BaseModel):
    """A saved credential for a backend provider/issuer."""

    name: str
    backend_issuer: CredentialIssuer
    is_default_for_issuer: bool
    submitted_time: datetime
    id: str

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "issuer": self.backend_issuer,
                "is_default_for_issuer": self.is_default_for_issuer,
                "created": self.submitted_time,
                "id": self.id,
            },
            index=[0],
        )


class Device(BaseModel):
    """A device in Nexus."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    backend_name: str
    device_name: Optional[str]
    nexus_hosted: bool
    stored_backend_info: StoredBackendInfo

    @property
    def backend_info(self) -> BackendInfo:
        """The BackendInfo for the Device."""
        return to_pytket_backend_info(self.stored_backend_info)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "backend_name": self.backend_name,
                "device_name": self.device_name,
                "nexus_hosted": self.nexus_hosted,
                "backend_info": to_pytket_backend_info(self.stored_backend_info),
            },
            index=[0],
        )

    @field_validator("backend_name")
    @classmethod
    def convert_backend_name(cls, v: str) -> str:
        """Convert the internal name for QuantinuumBackend."""
        if v == "EmulatorEnabledQuantinuumBackend":
            return "QuantinuumBackend"
        return v


class Quota(BaseModel):
    """A quota in Nexus."""

    name: str
    description: str
    usage: float
    quota: float | str

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(self.model_dump(), index=[0])


class Role(BaseModel):
    """A role for use in RBAC assignments."""

    id: UUID
    name: str
    description: str
    permissions: str
    type: Literal["role"] = "role"

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "description": self.description,
                "permissions": self.permissions,
                "id": self.id,
            },
            index=[0],
        )


class RoleInfo(BaseModel):
    """Information on a role assigned on a resource."""

    assignment_type: Literal["user", "team", "public"]
    assignee: TeamRef | UserRef | None
    role: Role

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""

        assignee_name: str | None = None
        match self.assignee:
            case TeamRef():
                assignee_name = self.assignee.name
            case UserRef():
                assignee_name = self.assignee.display_name
            case None:
                assignee_name = None
            case _:
                assert_never(self.assignee)

        return pd.DataFrame(
            {
                "assignment_type": self.assignment_type,
                "assignee": assignee_name,
                "role": self.role.name,
            },
            index=[0],
        )


class Property(BaseModel):
    """A property definition."""

    annotations: Annotations
    property_type: str
    required: bool
    color: str
    id: UUID

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "property_type": self.property_type,
                    "required": self.required,
                    "color": self.color,
                },
                index=[0],
            )
        )


class IssuerEnum(str, Enum):
    """Backend issuers in Nexus."""

    QUANTINUUM = "QUANTINUUM"
    IBMQ = "IBMQ"
    QISKIT = "QISKIT"
    BRAKET = "BRAKET"
    PROJECTQ = "PROJECTQ"
    QULACS = "QULACS"


def issuer_enum_to_config_str(issuer: IssuerEnum) -> list[str]:
    """Convert an IssuerEnum to a list of BackendConfig names."""

    match issuer:
        case IssuerEnum.QUANTINUUM:
            return ["QuantinuumConfig"]
        case IssuerEnum.IBMQ:
            return ["IBMQConfig"]
        case IssuerEnum.QISKIT:
            return [
                "IBMQEmulatorConfig",
                "AerConfig",
                "AerStateConfig",
                "AerUnitaryConfig",
            ]
        case IssuerEnum.PROJECTQ:
            return ["ProjectQConfig"]
        case IssuerEnum.QULACS:
            return ["QulacsConfig"]
        case IssuerEnum.BRAKET:
            return ["BraketConfig"]
        case _:
            assert_never(issuer)


def _register_to_pytket_node(register: Register) -> Node:
    """Convert a pytket Node object from a nexus-dataclasses Register object."""

    return Node.from_list(list(register))


def to_pytket_backend_info(backend: StoredBackendInfo) -> BackendInfo:
    """Reconstruct a pytket BackendInfo from the StoredBackendInfo instance."""

    stored_nodes = backend.device.nodes
    stored_edges = backend.device.edges
    architecture_edge_list = []

    # BackendInfo dictionary attributes are initialised as None,
    # Then dictionaries are built when they are to be populated.
    # This is to satisfy type-checking as well as BackendInfo expectations/tests.

    averaged_node_gate_errors = None
    averaged_readout_errors = None
    all_node_gate_errors = None
    all_readout_errors = None

    for stored_node in stored_nodes:
        # Create node from register
        new_pytket_node = _register_to_pytket_node(stored_node.unitid)

        # Build average node gate errors
        if stored_node.average_error is not None:
            if averaged_node_gate_errors is None:
                averaged_node_gate_errors = {}
            averaged_node_gate_errors[new_pytket_node] = stored_node.average_error
        # Build average readout errors
        if stored_node.readout_error is not None:
            if averaged_readout_errors is None:
                averaged_readout_errors = {}
            averaged_readout_errors[new_pytket_node] = stored_node.readout_error

        if stored_node.gate_errors:
            node_gate_errors = {
                getattr(OpType, optype_name): error
                for optype_name, error in stored_node.gate_errors.items()
            }
            if all_node_gate_errors is None:
                all_node_gate_errors = {}
            all_node_gate_errors[new_pytket_node] = node_gate_errors

        # Add stored_node readout errors to all_readout_errors
        stored_zero_state_readout_error = stored_node.zero_state_readout_error
        stored_one_state_readout_error = stored_node.one_state_readout_error
        if (
            stored_zero_state_readout_error is not None
            and stored_one_state_readout_error is not None
        ):
            if all_readout_errors is None:
                all_readout_errors = {}
            readout_errors = [
                [
                    1.0 - stored_zero_state_readout_error,
                    stored_zero_state_readout_error,
                ],
                [
                    stored_one_state_readout_error,
                    1.0 - stored_one_state_readout_error,
                ],
            ]
            all_readout_errors[new_pytket_node] = readout_errors

    # Build all_edge_gate_errors and averaged_edge_gate_errors from stored_edges
    all_edge_gate_errors = None
    averaged_edge_gate_errors = None

    for stored_edge in stored_edges:
        node_from = _register_to_pytket_node(stored_edge.unitid_from)
        node_to = _register_to_pytket_node(stored_edge.unitid_to)

        new_edge_tuple = (node_from, node_to)
        architecture_edge_list.append(new_edge_tuple)

        if stored_edge.average_error is not None:
            if averaged_edge_gate_errors is None:
                averaged_edge_gate_errors = {}
            averaged_edge_gate_errors[(node_from, node_to)] = stored_edge.average_error
        if stored_edge.gate_errors:
            edge_gate_errors = {
                getattr(OpType, optype_name): error
                for optype_name, error in stored_edge.gate_errors.items()
            }
            if all_edge_gate_errors is None:
                all_edge_gate_errors = {}
            all_edge_gate_errors[(node_from, node_to)] = edge_gate_errors

    architecture: Union[Architecture, FullyConnected] = (
        FullyConnected(
            backend.device.n_nodes
            if backend.device.n_nodes is not None
            else len(backend.device.nodes)
        )
        if backend.device.fully_connected
        else Architecture(architecture_edge_list)
    )

    gate_set = set()

    for gate in backend.gate_set:
        try:
            gate_set.add(getattr(OpType, gate))
        except AttributeError:
            logger.warning(
                "Unknown OpType in BackendInfo: `%`, will omit from BackendInfo."
                " Consider updating your pytket version."
            )

    return BackendInfo(
        name=backend.name,
        device_name=backend.device_name,
        version=backend.version,
        architecture=architecture,
        gate_set=gate_set,
        n_cl_reg=backend.n_cl_reg,
        supports_fast_feedforward=backend.supports_fast_feedforward,
        supports_reset=backend.supports_reset,
        supports_midcircuit_measurement=backend.supports_midcircuit_measurement,
        all_node_gate_errors=all_node_gate_errors,
        all_edge_gate_errors=all_edge_gate_errors,
        all_readout_errors=all_readout_errors,
        averaged_node_gate_errors=averaged_node_gate_errors,
        averaged_edge_gate_errors=averaged_edge_gate_errors,
        averaged_readout_errors=averaged_readout_errors,
        misc=backend.misc,
    )



"""Test basic functionality relating to the device module."""

import pytest
from pytket.backends.backendinfo import BackendInfo
from quantinuum_schemas.models.backend_config import BackendConfig
from quantinuum_schemas.models.backend_info import StoredBackendInfo

import qnexus as qnx
from qnexus.models import Device
from qnexus.models.references import DataframableList


def test_device_get_all(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get a list of devices."""
    devices = qnx.devices.get_all()
    assert isinstance(devices, DataframableList)

    for device in devices:
        assert isinstance(device, Device)
        assert isinstance(device.stored_backend_info, StoredBackendInfo)
        assert isinstance(device.backend_info, BackendInfo)


def test_supports_shots(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports shots."""

    supports_shots = qnx.devices.supports_shots(backend_config)
    assert isinstance(supports_shots, bool)


def test_supports_counts(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports counts."""

    supports_counts = qnx.devices.supports_counts(backend_config)
    assert isinstance(supports_counts, bool)


def test_supports_state(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports state."""

    supports_state = qnx.devices.supports_state(backend_config)
    assert isinstance(supports_state, bool)


def test_supports_unitary(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports unitary."""

    supports_unitary = qnx.devices.supports_unitary(backend_config)
    assert isinstance(supports_unitary, bool)


def test_supports_density_matrix(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports density matrix."""

    supports_density_matrix = qnx.devices.supports_density_matrix(backend_config)
    assert isinstance(supports_density_matrix, bool)


def test_supports_expectation(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports expectation."""

    supports_expectation = qnx.devices.supports_expectation(backend_config)
    assert isinstance(supports_expectation, bool)


def test_expectation_allows_nonhermitian(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports expectation_allows_nonhermitian."""

    expectation_allows_nonhermitian = qnx.devices.expectation_allows_nonhermitian(
        backend_config
    )
    assert isinstance(expectation_allows_nonhermitian, bool)


def test_supports_contextual_optimisation(
    _authenticated_nexus: None,
    backend_config: BackendConfig,
) -> None:
    """Test whether a BackendConfig supports contextual optimisation."""
    supports_contextual_optimisation = qnx.devices.supports_contextual_optimisation(
        backend_config
    )
    assert isinstance(supports_contextual_optimisation, bool)


def test_quantinuum_device_status(
    _authenticated_nexus: None,
) -> None:
    """Check that we can get the status of a hardware-hosted
    Quantinuum device."""

    unsupported_config = qnx.AerConfig()
    with pytest.raises(AssertionError):
        qnx.devices.status(unsupported_config)  # type: ignore

    backend_config = qnx.QuantinuumConfig(device_name="H1-1")

    machine_status = qnx.devices.status(backend_config)

    assert isinstance(machine_status, qnx.devices.DeviceStateEnum)



"""Test the token-based auth logic for the httpx Client.

N.B. these manipulate environment variables so currently run in isolation via scripts/run_unit_test.sh.
"""

from typing import Any, Generator
from uuid import uuid4

import httpx
import pytest
import respx

import qnexus as qnx
from qnexus.client import _nexus_client, get_nexus_client
from qnexus.client.utils import read_token, remove_token, write_token
from qnexus.config import CONFIG
from qnexus.exceptions import AuthenticationError


@pytest.fixture(autouse=True)
def clean_token_state() -> Generator[Any, Any, Any]:
    """Clean up token state before and after each test."""
    # Setup - clean token files
    remove_token("refresh_token")
    remove_token("access_token")
    global _nexus_client
    if _nexus_client is not None:
        _nexus_client.close()
        _nexus_client = None

    # Store tokens in a temporary location
    old_token_path = CONFIG.token_path
    CONFIG.token_path = f"/tmp/qnexus_tests/{str(uuid4())}"

    yield  # Run the test

    # Teardown - clean up after test
    remove_token("refresh_token")
    remove_token("access_token")
    get_nexus_client(reload=True)
    CONFIG.token_path = old_token_path


@respx.mock
def test_token_refresh() -> None:
    """Test the auth refresh logic, using in-memory token storage.

    Test that we can refresh the access token using the refresh token
    and that the new access token is used in subsequent requests.
    """

    old_id_token = "dummy_id"
    refreshed_access_token = "new_dummy_id"

    write_token("refresh_token", "dummy_oat")
    write_token("access_token", old_id_token)

    # Mock the list projects endpoint to force a refresh
    list_project_route = respx.get(
        f"{get_nexus_client().base_url}/api/projects/v1beta"
    ).mock(
        side_effect=[
            httpx.Response(401),
            httpx.Response(200, json={"included": {}, "data": []}),
        ]
    )

    # Mock the refresh endpoint
    refresh_token_route = respx.post(
        f"{get_nexus_client().base_url}/auth/tokens/refresh"
    ).mock(
        return_value=httpx.Response(
            200,
            headers={
                "set-cookie": f"myqos_id={refreshed_access_token}; "
                "HttpOnly; Path=/; SameSite=Lax; Secure"
            },
        )
    )

    qnx.projects.get_all().list()

    assert list_project_route.called
    assert refresh_token_route.called

    # Confirm that the access token was updated
    assert read_token("access_token") == refreshed_access_token
    assert get_nexus_client().auth.cookies.get("myqos_id") == refreshed_access_token  # type: ignore

    # confirm that the request headers were updated
    first_cookie_header = list_project_route.calls[0].request.headers["cookie"]
    assert f"myqos_id={old_id_token}" in first_cookie_header

    last_cookie_header = list_project_route.calls[-1].request.headers["cookie"]
    assert f"myqos_id={refreshed_access_token}" in last_cookie_header


@respx.mock
def test_token_refresh_expired() -> None:
    """Test the case of an expired refresh token, using in-memory token storage."""

    write_token("refresh_token", "dummy_oat")
    write_token("access_token", "dummy_id")

    # Mock the list projects endpoint to force a refresh
    list_project_route = respx.get(
        f"{get_nexus_client().base_url}/api/projects/v1beta"
    ).mock(return_value=httpx.Response(401))

    # Mock the expiry of the refresh token
    refresh_token_route = respx.post(
        f"{get_nexus_client().base_url}/auth/tokens/refresh"
    ).mock(return_value=httpx.Response(401))

    with pytest.raises(AuthenticationError):
        qnx.projects.get_all().list()

    assert list_project_route.called
    assert refresh_token_route.called


def test_nexus_client_reloads_tokens() -> None:
    """Test the reload functionality of the nexus client.

    Test that if we write new tokens and reload the client,
    that the new tokens are used."""

    oat_one = "dummy_oat_one"
    oat_two = "dummy_oat_two"

    write_token("refresh_token", oat_one)
    client_one = get_nexus_client(reload=True)
    assert client_one.auth.cookies.get("myqos_oat") == oat_one  # type: ignore

    write_token("refresh_token", oat_two)
    client_two = get_nexus_client()
    assert client_two.auth.cookies.get("myqos_oat") == oat_one  # type: ignore

    client_two = get_nexus_client(reload=True)
    assert client_two.auth.cookies.get("myqos_oat") == oat_two  # type: ignore


def test_nexus_client_reloads_domain() -> None:
    """Test the reload functionality of the nexus client.
    We should be able to change the domain in the config
    and have the client reload with the new domain obtainable
    via a getter function.
    """

    domain_one = "dummy_domain_one.com"
    domain_two = "dummy_domain_two.com"

    CONFIG.domain = domain_one
    # mock login
    write_token("refresh_token", "dummy_oat")
    write_token("access_token", "dummy_id")

    client_one = get_nexus_client(reload=True)
    client_two = get_nexus_client(reload=True)

    assert domain_one in str(client_one.base_url)
    assert domain_one in str(client_two.base_url)

    CONFIG.domain = domain_two

    assert domain_two not in str(client_one.base_url)
    assert domain_two not in str(client_two.base_url)

    client_two = get_nexus_client(reload=True)
    # client_two getter should not effect client_one
    assert domain_two not in str(client_one.base_url)
    # client_two getter should reload the client
    assert domain_two in str(client_two.base_url)



"""Tests that we can use all available backend configs."""

from datetime import datetime

from pytket.backends.backendresult import BackendResult
from pytket.backends.status import StatusEnum
from pytket.circuit import Circuit

import qnexus as qnx

CONFIGS_REQUIRE_NO_MEASURE = [qnx.AerUnitaryConfig]
CONFIGS_NOT_TO_EXECUTE = [
    # We won't execute for these configs as they run on quantum hardware
    qnx.IBMQConfig
]


def test_basic_backend_config_usage(
    _authenticated_nexus: None,
    backend_config: qnx.BackendConfig,
    qa_project_name: str,
) -> None:
    """Test basic functionality of supported BackendConfigs."""

    project_ref = qnx.projects.get(name_like=qa_project_name)

    my_circ = Circuit(2, 2).H(0).CX(0, 1)

    if backend_config.__class__ not in CONFIGS_REQUIRE_NO_MEASURE:
        my_circ.measure_all()

    my_circ = qnx.circuits.upload(
        circuit=my_circ,
        name=f"QA BackendConfig circuit {backend_config.__class__.__name__}_{datetime.now()}",
        description=f"This can be safely deleted. Test Run: {datetime.now()}",
        project=project_ref,
    )

    compile_job_ref = qnx.start_compile_job(
        programs=[my_circ],
        name=f"QA BackendConfig compile job {backend_config.__class__.__name__}_{datetime.now()}",
        optimisation_level=2,
        backend_config=backend_config,
        project=project_ref,
    )

    qnx.jobs.wait_for(compile_job_ref)

    # Check the backend config as stored in Nexus matches the one specified
    # QuantinuumConfig noisy_simulation flag getting reset in Nexus
    assert backend_config.model_dump(exclude={"noisy_simulation"}) == qnx.jobs.get(
        id=compile_job_ref.id
    ).backend_config.model_dump(exclude={"noisy_simulation"})

    execute_job_ref = qnx.start_execute_job(
        programs=[item.get_output() for item in qnx.jobs.results(compile_job_ref)],
        name=f"Execute {backend_config.__class__.__name__}_{datetime.now()}",
        n_shots=[100],
        backend_config=backend_config,
        project=project_ref,
    )

    if backend_config.__class__ in CONFIGS_NOT_TO_EXECUTE:
        qnx.jobs.wait_for(execute_job_ref, wait_for_status=StatusEnum.QUEUED)
        qnx.jobs.cancel(execute_job_ref)
        qnx.jobs.wait_for(execute_job_ref, wait_for_status=StatusEnum.CANCELLED)
        return

    qnx.jobs.wait_for(execute_job_ref)

    execute_job_result_refs = qnx.jobs.results(execute_job_ref)

    for result_ref in execute_job_result_refs:
        assert isinstance(result_ref.download_result(), BackendResult)



"""Test basic functionality relating to the project module."""

from datetime import datetime

import pandas as pd
import pytest

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.references import ProjectRef


def test_project_get(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can get a specific unique project
    by name or id, or get an appropriate exception."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    assert isinstance(my_proj, ProjectRef)

    my_proj_2 = qnx.projects.get(id=my_proj.id)
    assert my_proj == my_proj_2

    with pytest.raises(qnx_exc.NoUniqueMatch):
        qnx.projects.get()

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.projects.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_project_get_all(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get an iterator over all projects."""

    my_proj_db_matches = qnx.projects.get_all()

    assert isinstance(my_proj_db_matches.count(), int)
    assert isinstance(my_proj_db_matches.summarize(), pd.DataFrame)

    assert isinstance(next(my_proj_db_matches), ProjectRef)


def test_project_create(
    _authenticated_nexus: None,
) -> None:
    """Test that we can create a project and add a property definition."""

    project_name = f"QA_test_project_create_{datetime.now()}"

    my_new_project = qnx.projects.create(name=project_name)

    assert isinstance(my_new_project, ProjectRef)

    test_property_name = "QA_test_prop"

    qnx.projects.add_property(
        name="QA_test_prop",
        property_type="bool",
        project=my_new_project,
        description="A test property for my QA test project",
    )

    test_props = qnx.projects.get_properties(my_new_project)

    assert len(test_props) == 1
    assert test_props[0].annotations.name == test_property_name

    with pytest.raises(qnx_exc.ResourceDeleteFailed):
        qnx.projects.delete(my_new_project)

    qnx.projects.update(my_new_project, archive=True)

    qnx.projects.delete(my_new_project)

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.projects.get(name_like=project_name)


def test_project_get_or_create(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get or create a project."""

    project_name = f"QA_test_project_get_or_create_{datetime.now()}"

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.projects.get(name_like=project_name)

    my_new_project = qnx.projects.get_or_create(name=project_name)

    assert isinstance(my_new_project, ProjectRef)

    my_new_project_again = qnx.projects.get_or_create(name=project_name)

    assert my_new_project == my_new_project_again

    qnx.projects.update(my_new_project, archive=True)
    qnx.projects.delete(my_new_project)


def test_project_summarize(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can get summary information on the state of a project."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    assert isinstance(my_proj, ProjectRef)

    project_summary = qnx.projects.summarize(my_proj)

    assert isinstance(project_summary, pd.DataFrame)

    assert (project_summary["total_jobs"] > 0).all()



"""Client API for circuits in Nexus."""

from datetime import datetime
from typing import Any, Union, cast
from uuid import UUID

from httpx import QueryParams
from pytket.circuit import Circuit
from pytket.utils.serialization.migration import circuit_dict_from_pytket1_dict
from quantinuum_schemas.models.backend_config import BackendConfig, QuantinuumConfig

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
)
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    FuzzyNameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import CircuitRef, DataframableList, ProjectRef


class Params(
    SortFilter,
    PaginationFilter,
    FuzzyNameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
    ScopeFilter,
):
    """Params for filtering circuits."""


@merge_project_from_context
def get_all(
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[CircuitRef]:
    """Get a NexusIterator over circuits with optional filters."""

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="Circuit",
        nexus_url="/api/circuits/v1beta",
        params=params,
        wrapper_method=_to_circuitref,
        nexus_client=get_nexus_client(),
    )


def _to_circuitref(page_json: dict[str, Any]) -> DataframableList[CircuitRef]:
    """Convert JSON response dict to a list of CircuitRefs."""

    circuit_refs: DataframableList[CircuitRef] = DataframableList([])

    for circuit_data in page_json["data"]:
        project_id = circuit_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        circuit_refs.append(
            CircuitRef(
                id=UUID(circuit_data["id"]),
                annotations=Annotations.from_dict(circuit_data["attributes"]),
                project=project,
            )
        )
    return circuit_refs


def get(
    *,
    id: Union[UUID, str, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> CircuitRef:
    """
    Get a single circuit using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(circuit_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    circuit: Circuit,
    project: ProjectRef | None = None,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> CircuitRef:
    """Upload a pytket Circuit to Nexus."""
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    circuit_dict = circuit.to_dict()
    circuit_name = name if name else circuit.name
    if circuit_name is None:
        raise ValueError("Circuit must have a name to be uploaded")

    annotations = CreateAnnotations(
        name=circuit_name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    circuit_dict.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": circuit_dict,
            "relationships": relationships,
            "type": "circuit",
        }
    }

    res = get_nexus_client().post("/api/circuits/v1beta", json=req_dict)

    # https://cqc.atlassian.net/browse/MUS-3054
    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return CircuitRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: CircuitRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> CircuitRef:
    """Update the annotations on a CircuitRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "circuit",
        }
    }

    res = get_nexus_client().patch(f"/api/circuits/v1beta/{ref.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return CircuitRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


def _fetch_by_id(circuit_id: UUID | str, scope: ScopeFilterEnum | None) -> CircuitRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/circuits/v1beta/{circuit_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return CircuitRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


def _fetch_circuit(handle: CircuitRef) -> Circuit:
    """Utility method for fetching a pytket circuit from a CircuitRef."""
    res = get_nexus_client().get(f"/api/circuits/v1beta/{handle.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_data_attributes_dict = res.json()["data"]["attributes"]
    circuit_dict = {k: v for k, v in res_data_attributes_dict.items() if v is not None}

    return Circuit.from_dict(circuit_dict_from_pytket1_dict(circuit_dict))


def cost(
    circuit_ref: CircuitRef,
    n_shots: int,
    backend_config: BackendConfig,
    syntax_checker: str | None = None,
) -> float | None:
    """Calculate the cost (in HQC) of running a circuit for n_shots
    number of shots on a specific Quantinuum device."""

    if not isinstance(backend_config, QuantinuumConfig):
        raise ValueError("QuantinuumConfig is the only supported backend config")

    params = {
        "n_shots": n_shots,
        "device_name": backend_config.device_name,
    }
    if syntax_checker:
        params["syntax_checker"] = syntax_checker

    res = get_nexus_client().get(
        f"/api/circuits/v1beta/cost/{circuit_ref.id}",
        params=cast(QueryParams, params),
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    circuit_cost: float | None = res.json()
    return circuit_cost



"""Test basic functionality relating to the role module."""

from datetime import datetime

import pandas as pd

import qnexus as qnx
from qnexus.config import CONFIG
from qnexus.models import Role
from qnexus.models.references import TeamRef, UserRef


def test_role_get(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get a specific Role."""

    role = qnx.roles.get(name="Administrator")
    assert isinstance(role, Role)


def test_role_get_all(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get all assignment role definitions."""

    all_roles = qnx.roles.get_all()

    assert len(all_roles) == 4
    assert isinstance(all_roles.df(), pd.DataFrame)
    assert isinstance(all_roles[0], Role)


def test_team_assignment(
    _authenticated_nexus: None,
    qa_team_name: str,
) -> None:
    """Test that we can get all assignment role definitions."""

    new_project_ref = qnx.projects.create(name=f"QA_test_project_{datetime.now()}_0")

    team = qnx.teams.get(name=qa_team_name)

    qnx.roles.assign_team(resource_ref=new_project_ref, team=team, role="Administrator")

    assignments = qnx.roles.assignments(resource_ref=new_project_ref)

    assert len(assignments) == 2

    team_assignment = next(
        assign for assign in assignments if assign.assignment_type == "team"
    )
    assert team_assignment.assignment_type == "team"
    assert isinstance(team_assignment.assignee, TeamRef)
    assert team_assignment.assignee.id == team.id

    qnx.projects.update(new_project_ref, archive=True)
    qnx.projects.delete(new_project_ref)


def test_user_assignment(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get all assignment role definitions."""

    new_project_ref = qnx.projects.create(name=f"QA_test_project_{datetime.now()}_1")

    qnx.roles.assign_user(
        resource_ref=new_project_ref,
        user_email=CONFIG.qa_user_email,
        role="Contributor",
    )

    assignments = qnx.roles.assignments(resource_ref=new_project_ref)

    assert len(assignments) == 2

    contrib_assignment = next(
        assign for assign in assignments if assign.role.name == "Contributor"
    )
    assert contrib_assignment.assignment_type == "user"
    assert isinstance(contrib_assignment.assignee, UserRef)

    admin_assignment = next(
        assign for assign in assignments if assign.role.name == "Administrator"
    )
    assert admin_assignment.assignment_type == "user"
    assert isinstance(admin_assignment.assignee, UserRef)

    qnx.projects.update(new_project_ref, archive=True)
    qnx.projects.delete(new_project_ref)



"""CLI for qnexus."""

import os
from typing import Any, Callable

from click import Command, Option


def is_documented_by(original: Callable) -> Callable:  # type: ignore
    """Forward the documentation from the underlying client function."""

    def wrapper(target: Callable) -> Callable:  # type: ignore
        target.__doc__ = original.__doc__
        return target

    return wrapper


current_path = os.getcwd()
current_dir = current_path.split(os.sep)[-1]


# # QNX utils interface
# @click.command()
# @is_documented_by(_status)
# def status():
#     click.echo(_status())


# @click.command()
# def init():
#     """Initialize a new qnexus project."""
#     if config:
#         raise click.ClickException(
#             Fore.GREEN + f"Project already initialized: {Fore.YELLOW}"
#         )
#     if config is None:
#         name: str = click.prompt(
#             "Enter a project name:",
#             default=current_dir,
#         )
#         click.echo(Fore.GREEN + f"Intialized qnexus project: {name}")


def add_options_to_command(command: Command, model: Any) -> None:
    """Add click options using fields of a pydantic model."""
    # Annotate command with options from dict
    for field, value in model.model_fields.items():
        command.params.append(
            Option(
                [f"--{field}"],
                help=value.description,
                show_default=True,
                default=value.default,
                multiple=isinstance(value.default, list),
            )
        )



"""Basic checks for HUGR functionality."""

from pathlib import Path

from hugr.package import Package

import qnexus as qnx


def test_hugr_encode_decode() -> None:
    """
    Given a known-good binary-encoded HUGR in a file, test that it can be loaded
    into a hugr.package.Package instance, then serialised and deserialised with
    fidelity.

    We expect this test to continue passing when client.hugr.ENVELOPE_CONFIG is
    changed.
    """
    with open(Path(__file__).parent.resolve() / "data" / "hugr.dat", "rb") as fp:
        data = fp.read()
        package = Package.from_bytes(data)

    encoded_package = qnx.hugr._encode_hugr(package)
    decoded_package = qnx.hugr._decode_hugr(encoded_package)
    assert encoded_package == qnx.hugr._encode_hugr(decoded_package)



"""Client API for execution in Nexus."""

from typing import Union, cast

from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.backends.status import StatusEnum

import qnexus.exceptions as qnx_exc
from qnexus.client import circuits as circuit_api
from qnexus.client import get_nexus_client
from qnexus.client import hugr as hugr_api
from qnexus.client import qir as qir_api
from qnexus.client.utils import accept_circuits_for_programs
from qnexus.context import get_active_project, merge_properties_from_context
from qnexus.models import BackendConfig, StoredBackendInfo, to_pytket_backend_info
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResultRef,
    HUGRRef,
    JobType,
    ProjectRef,
    QIRRef,
    ResultType,
    WasmModuleRef,
)
from qnexus.models.utils import assert_never


@accept_circuits_for_programs
@merge_properties_from_context
def start_execute_job(
    programs: Union[ExecutionProgram, list[ExecutionProgram]],
    n_shots: list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    postprocess: bool = False,
    noisy_simulator: bool = True,
    language: Language = Language.AUTO,
    seed: int | None = None,
    credential_name: str | None = None,
    wasm_module: WasmModuleRef | None = None,
    user_group: str | None = None,
) -> ExecuteJobRef:
    """
    Submit an execute job to be run in Nexus. Returns an ``ExecuteJobRef``
    object which can be used to check the job's status.  See ``qnexus.execute``
    for a utility method that waits for the results and returns them.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    program_ids = (
        [str(p.id) for p in programs]
        if isinstance(programs, list)
        else [str(programs.id)]
    )

    if len(n_shots) != len(program_ids):
        raise ValueError("Number of programs must equal number of n_shots.")

    attributes_dict = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes_dict.update(
        {
            "job_type": "execute",
            "definition": {
                "job_definition_type": "execute_job_definition",
                "backend_config": backend_config.model_dump(exclude_none=True),
                "user_group": user_group,
                "valid_check": valid_check,
                "postprocess": postprocess,
                "noisy_simulator": noisy_simulator,
                "language": (
                    language.value if isinstance(language, Language) else language
                ),
                "seed": seed,
                "wasm_module_id": str(wasm_module.id) if wasm_module else None,
                "credential_name": credential_name,
                "items": [
                    {"program_id": program_id, "n_shots": n_shot}
                    for program_id, n_shot in zip(program_ids, n_shots)
                ],
            },
        }
    )
    relationships = {
        "project": {"data": {"id": str(project.id), "type": "project"}},
    }
    req_dict = {
        "data": {
            "attributes": attributes_dict,
            "relationships": relationships,
            "type": "job",
        }
    }

    resp = get_nexus_client().post(
        "/api/jobs/v1beta3",
        json=req_dict,
    )
    if resp.status_code != 202:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    return ExecuteJobRef(
        id=resp.json()["data"]["id"],
        annotations=Annotations.from_dict(resp.json()["data"]["attributes"]),
        job_type=JobType.EXECUTE,
        last_status=StatusEnum.SUBMITTED,
        last_message="",
        project=project,
        backend_config_store=backend_config,
    )


def _results(
    execute_job: ExecuteJobRef,
    allow_incomplete: bool = False,
) -> DataframableList[ExecutionResultRef]:
    """Get the results from an execute job."""

    resp = get_nexus_client().get(f"/api/jobs/v1beta3/{execute_job.id}")

    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]["status"]

    if job_status != "COMPLETED" and not allow_incomplete:
        raise qnx_exc.ResourceFetchFailed(message=f"Job status: {job_status}")

    execute_results: DataframableList[ExecutionResultRef] = DataframableList([])

    for item in resp_data["attributes"]["definition"]["items"]:
        # Check if item is in a "final" state and append to results
        if item["status"]["status"] in (
            "CANCELLED",
            "ERROR",
            "DEPLETED",
            "TERMINATED",
            "COMPLETED",
        ):
            result_type: ResultType

            match item["result_type"]:
                case ResultType.QSYS:
                    result_type = ResultType.QSYS
                case ResultType.PYTKET:
                    result_type = ResultType.PYTKET
                case _:
                    assert_never(item["result_type"])

            result_ref = ExecutionResultRef(
                id=item["result_id"],
                annotations=execute_job.annotations,
                project=execute_job.project,
                result_type=result_type,
            )

            execute_results.append(result_ref)

    return execute_results


def _fetch_pytket_execution_result(
    result_ref: ExecutionResultRef,
) -> tuple[BackendResult, BackendInfo, Union[CircuitRef, QIRRef]]:
    """Get the results for an execute job item."""
    assert result_ref.result_type == ResultType.PYTKET, "Incorrect result type"

    res = get_nexus_client().get(f"/api/results/v1beta3/{result_ref.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()

    program_data = res_dict["data"]["relationships"]["program"]["data"]
    program_id = program_data["id"]
    program_type = program_data["type"]

    input_program: Union[CircuitRef | QIRRef]
    match program_type:
        case "circuit":
            input_program = circuit_api._fetch_by_id(program_id, scope=None)
        case "qir":
            input_program = qir_api._fetch_by_id(program_id, scope=None)
        case _:
            raise ValueError(f"Unknown program type {type}")

    results_data = res_dict["data"]["attributes"]

    results_dict = {k: v for k, v in results_data.items() if v != [] and v is not None}

    backend_result = BackendResult.from_dict(results_dict)

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    return (backend_result, backend_info, input_program)


def _fetch_qsys_execution_result(
    result_ref: ExecutionResultRef,
) -> tuple[QsysResult, BackendInfo, HUGRRef | QIRRef]:
    """Get the results of a next-gen Qsys execute job."""
    assert result_ref.result_type == ResultType.QSYS, "Incorrect result type"

    res = get_nexus_client().get(f"/api/qsys_results/v1beta/{result_ref.id}")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()

    input_program_id = res_dict["data"]["relationships"]["program"]["data"]["id"]

    input_program: HUGRRef | QIRRef
    match res_dict["data"]["relationships"]["program"]["data"]["type"]:
        case "hugr":
            input_program = hugr_api._fetch_by_id(
                input_program_id,
                scope=None,
            )
        case "qir":
            input_program = qir_api._fetch_by_id(
                input_program_id,
                scope=None,
            )

    qsys_result = QsysResult(res_dict["data"]["attributes"]["results"])

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    return (
        qsys_result,
        backend_info,
        input_program,
    )



"""Test basic functionality relating to the circuit module."""

from datetime import datetime

import pandas as pd
import pytest
from pytket.circuit import Circuit

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import CircuitRef


def test_circuit_get(
    _authenticated_nexus: None,
    qa_circuit_name: str,
) -> None:
    """Test that we can get a specific unique CircuitRef
    by name or id, or get an appropriate exception."""

    my_circ = qnx.circuits.get(name_like=qa_circuit_name)
    assert isinstance(my_circ, CircuitRef)

    assert isinstance(my_circ.download_circuit(), Circuit)

    my_circ_2 = qnx.circuits.get(id=my_circ.id)
    # For some reason direct equality check fails
    assert my_circ.id == my_circ_2.id
    assert my_circ.annotations == my_circ_2.annotations
    assert my_circ.project == my_circ_2.project

    with pytest.raises(qnx_exc.NoUniqueMatch):
        qnx.circuits.get()

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.circuits.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_circuit_get_all(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get an iterator over all circuits."""

    my_circ_db_matches = qnx.circuits.get_all()

    assert isinstance(my_circ_db_matches.count(), int)
    assert isinstance(my_circ_db_matches.summarize(), pd.DataFrame)

    assert isinstance(next(my_circ_db_matches), CircuitRef)


def test_circuit_create(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can create a circuit and add a property value."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    circuit_name = f"QA_test_circuit_{datetime.now()}"

    my_circ = Circuit(2, 2).H(0).CX(0, 1).measure_all()

    my_new_circuit = qnx.circuits.upload(
        circuit=my_circ, name=circuit_name, project=my_proj
    )

    assert isinstance(my_new_circuit, CircuitRef)

    test_property_name = "QA_test_prop"
    test_prop_value = "foo"

    updated_circuit_ref = qnx.circuits.update(
        ref=my_new_circuit,
        properties=PropertiesDict({test_property_name: test_prop_value}),
    )

    assert (
        updated_circuit_ref.annotations.properties[test_property_name]
        == test_prop_value
    )


def test_circuit_get_cost(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can get the cost to run a CircuitRef,
    on a particular Quantinuum Systems device."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    my_q_systems_circuit = qnx.circuits.upload(
        circuit=Circuit(2, 2).ZZPhase(0.5, 0, 1).measure_all(),
        name="qa_q_systems_circuit",
        project=my_proj,
    )

    cost = qnx.circuits.cost(
        circuit_ref=my_q_systems_circuit,
        n_shots=10,
        backend_config=qnx.QuantinuumConfig(device_name="H1-1E"),
        syntax_checker="H1-1SC",
    )

    assert isinstance(cost, float)
    assert cost > 0.0



"""Test basic functionality relating to the team module."""

from datetime import datetime

import pandas as pd

import qnexus as qnx
from qnexus.models.references import TeamRef


def test_team_get(_authenticated_nexus: None, qa_team_name: str) -> None:
    """Test that we can get a specific unique TeamRef."""

    my_team = qnx.teams.get(name=qa_team_name)
    assert isinstance(my_team, TeamRef)


def test_team_get_all(
    _authenticated_nexus: None,
) -> None:
    """Test that we can get all teams (currently not a NexusIterator)."""

    my_teams = qnx.teams.get_all()

    assert len(my_teams) > 1
    assert isinstance(my_teams.df(), pd.DataFrame)
    assert isinstance(my_teams[0], TeamRef)


def test_team_create(
    _authenticated_nexus: None,
) -> None:
    """Test that we can create a team."""

    team_name = f"qa_test_team_{datetime.now()}"
    team_description = "A test team (can be deleted)"

    my_new_team = qnx.teams.create(name=team_name, description=team_description)

    assert isinstance(my_new_team, TeamRef)
    assert my_new_team.name == team_name
    assert my_new_team.description == team_description



"""Test filesystem operations with Nexus Refs."""

import shutil
from datetime import datetime
from pathlib import Path
from typing import cast

import qnexus as qnx
from qnexus.filesystem import load, save
from qnexus.models.references import (
    BaseRef,
    CompileJobRef,
    ExecuteJobRef,
    JobRef,
)


def test_save_load(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
    qa_team_name: str,
    qa_compile_job_name: str,
    qa_execute_job_name: str,
    qa_wasm_module_name: str,
    qa_hugr_name: str,
    qa_qir_name: str,
) -> None:
    """Test Ref serialization and deserialization roundtrip,
    and check that all Ref types are covered."""

    user_ref = qnx.users.get_self()
    project_ref = qnx.projects.get(name_like=qa_project_name)
    team_ref = qnx.teams.get(name=qa_team_name)
    circuit_ref = qnx.circuits.get(name_like=qa_circuit_name)
    execute_job_ref = cast(ExecuteJobRef, qnx.jobs.get(name_like=qa_execute_job_name))
    compile_job_ref = cast(CompileJobRef, qnx.jobs.get(name_like=qa_compile_job_name))
    execute_result_ref = qnx.jobs.results(execute_job_ref)[0]
    compile_result_ref = qnx.jobs.results(compile_job_ref)[0]
    # Copy to avoid adding cached data to the ref which will
    # affect the equality assertion below
    compilation_pass_ref = compile_result_ref.model_copy().get_passes()[0]
    wasm_module_ref = qnx.wasm_modules.get(name_like=qa_wasm_module_name)
    hugr_ref = qnx.hugr.get(name_like=qa_hugr_name)
    qir_ref = qnx.qir.get(name_like=qa_qir_name)

    test_refs = [
        user_ref,
        project_ref,
        team_ref,
        circuit_ref,
        execute_job_ref,
        compile_job_ref,
        execute_result_ref,
        compile_result_ref,
        compilation_pass_ref,
        wasm_module_ref,
        hugr_ref,
        qir_ref,
    ]
    test_ref_types = set(type(test_ref) for test_ref in test_refs)
    all_refs = set(BaseRef.__subclasses__())
    # remove JobRef as its only used as a base class
    all_refs.remove(JobRef)  # type: ignore
    assert test_ref_types == all_refs

    test_ref_path = Path().cwd() / f"refs_test_{datetime.now()}"

    save(ref=project_ref, path=test_ref_path / "project", mkdir=True)
    save(ref=team_ref, path=test_ref_path / "team", mkdir=True)
    save(ref=circuit_ref, path=test_ref_path / "circuit", mkdir=True)
    save(ref=execute_job_ref, path=test_ref_path / "execute", mkdir=True)
    save(ref=compile_job_ref, path=test_ref_path / "compile", mkdir=True)
    save(ref=execute_result_ref, path=test_ref_path / "execute_result", mkdir=True)
    save(ref=compile_result_ref, path=test_ref_path / "compile_result", mkdir=True)
    save(ref=compilation_pass_ref, path=test_ref_path / "compilation_pass", mkdir=True)
    save(ref=user_ref, path=test_ref_path / "user", mkdir=True)
    save(ref=wasm_module_ref, path=test_ref_path / "wasm_module", mkdir=True)
    save(ref=hugr_ref, path=test_ref_path / "hugr_ref", mkdir=True)

    project_ref_2 = load(path=test_ref_path / "project")
    team_ref_2 = load(path=test_ref_path / "team")
    circuit_ref_2 = load(path=test_ref_path / "circuit")
    execute_job_ref_2 = load(path=test_ref_path / "execute")
    compile_job_ref_2 = load(path=test_ref_path / "compile")
    execute_result_ref_2 = load(path=test_ref_path / "execute_result")
    compile_result_ref_2 = load(path=test_ref_path / "compile_result")
    compilation_pass_ref_2 = load(path=test_ref_path / "compilation_pass")
    user_ref_2 = load(path=test_ref_path / "user")
    wasm_module_ref_2 = load(path=test_ref_path / "wasm_module")
    hugr_ref_2 = load(path=test_ref_path / "hugr_ref")

    assert project_ref == project_ref_2
    assert team_ref == team_ref_2
    assert circuit_ref == circuit_ref_2
    assert execute_job_ref == execute_job_ref_2
    assert compile_job_ref == compile_job_ref_2
    assert execute_result_ref == execute_result_ref_2
    assert compile_result_ref == compile_result_ref_2
    assert compilation_pass_ref == compilation_pass_ref_2
    assert user_ref == user_ref_2
    assert wasm_module_ref == wasm_module_ref_2
    assert hugr_ref == hugr_ref_2

    # Clean up the saved ref files
    shutil.rmtree(test_ref_path)



"""Test basic functionality relating to the job module."""

from collections import Counter
from datetime import datetime
from time import sleep

import pandas as pd
import pytest
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Circuit
from quantinuum_schemas.models.backend_config import BaseBackendConfig
from quantinuum_schemas.models.hypertket_config import HyperTketConfig

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.references import (
    CircuitRef,
    CompilationPassRef,
    CompilationResultRef,
    CompileJobRef,
    ExecuteJobRef,
    JobRef,
)


def test_job_get_all(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can get an iterator over all jobs."""

    project_ref = qnx.projects.get(name_like=qa_project_name)

    my_job_db_matches = qnx.jobs.get_all(project=project_ref)

    assert isinstance(my_job_db_matches.count(), int)
    assert isinstance(my_job_db_matches.summarize(), pd.DataFrame)

    for job_ref in my_job_db_matches.list():
        assert isinstance(job_ref, JobRef)
        assert job_ref.backend_config is not None
        assert isinstance(job_ref.backend_config, BaseBackendConfig)


def test_job_get_by_id(
    _authenticated_nexus: None,
    qa_compile_job_name: str,
    qa_execute_job_name: str,
) -> None:
    """Test that we can get JobRefs by id."""

    my_compile_job = qnx.jobs.get(name_like=qa_compile_job_name)
    assert isinstance(my_compile_job, CompileJobRef)
    assert isinstance(my_compile_job.backend_config, BaseBackendConfig)

    my_compile_job_2 = qnx.jobs.get(id=my_compile_job.id)
    assert my_compile_job == my_compile_job_2

    my_execute_job = qnx.jobs.get(name_like=qa_execute_job_name)
    assert isinstance(my_execute_job, ExecuteJobRef)
    assert isinstance(my_execute_job.backend_config, BaseBackendConfig)

    my_execute_job_2 = qnx.jobs.get(id=my_execute_job.id)
    assert my_execute_job == my_execute_job_2


def test_compile_job_get(
    _authenticated_nexus: None,
    qa_compile_job_name: str,
) -> None:
    """Test that we can get a specific unique CompileJobRef,
    or get an appropriate exception."""

    my_job = qnx.jobs.get(name_like=qa_compile_job_name)
    assert isinstance(my_job, CompileJobRef)

    with pytest.raises(qnx_exc.NoUniqueMatch):
        qnx.jobs.get()

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.jobs.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_execute_job_get(
    _authenticated_nexus: None,
    qa_execute_job_name: str,
) -> None:
    """Test that we can get a specific unique ExecuteJobRef,
    or get an appropriate exception."""

    my_job = qnx.jobs.get(name_like=qa_execute_job_name)
    assert isinstance(my_job, ExecuteJobRef)

    with pytest.raises(qnx_exc.NoUniqueMatch):
        qnx.jobs.get()

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.jobs.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_submit_compile(
    _authenticated_nexus_circuit_ref: CircuitRef,
    qa_project_name: str,
) -> None:
    """Test that we can run a compile job in Nexus, wait for the job to complete and
    obtain the results from the compilation."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    config = qnx.AerConfig()

    compile_job_ref = qnx.start_compile_job(
        programs=[_authenticated_nexus_circuit_ref],
        name=f"qnexus_integration_test_compile_job_{datetime.now()}",
        project=my_proj,
        backend_config=config,
        skip_intermediate_circuits=False,
    )

    assert isinstance(compile_job_ref, CompileJobRef)

    qnx.jobs.wait_for(compile_job_ref)

    compile_results = qnx.jobs.results(compile_job_ref)

    assert len(compile_results) == 1
    assert isinstance(compile_results[0], CompilationResultRef)

    assert isinstance(compile_results[0].get_input(), CircuitRef)
    assert isinstance(compile_results[0].get_output(), CircuitRef)

    first_pass_data = compile_results[0].get_passes()[0]

    assert isinstance(first_pass_data, CompilationPassRef)
    assert isinstance(first_pass_data.get_input(), CircuitRef)
    assert isinstance(first_pass_data.get_output(), CircuitRef)
    assert isinstance(first_pass_data.pass_name, str)

    cj_ref = qnx.jobs.get(id=compile_job_ref.id)
    assert cj_ref.backend_config == config


def test_compile(
    _authenticated_nexus_circuit_ref: CircuitRef,
    qa_project_name: str,
) -> None:
    """Test that we can run the utility compile function and get compiled circuits."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    compiled_circuits = qnx.compile(
        programs=[_authenticated_nexus_circuit_ref],
        name=f"qnexus_integration_test_compile_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
    )

    assert len(compiled_circuits) == 1
    assert isinstance(compiled_circuits[0], CircuitRef)


def test_get_results_for_incomplete_compile(
    _authenticated_nexus_circuit_ref: CircuitRef,
    qa_project_name: str,
) -> None:
    """Test that we can run a compile job in Nexus, and fetch complete results for
    an otherwise errored/incomplete job."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    compile_job_ref = qnx.start_compile_job(
        programs=[_authenticated_nexus_circuit_ref],
        name=f"qnexus_integration_test_compile_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
    )
    # Unsure how to guarantee that the job will be incomplete
    # This check may fail if the compilation completes quickly
    compile_results = qnx.jobs.results(compile_job_ref, allow_incomplete=True)
    assert len(compile_results) == 0


def test_compile_hypertket(
    _authenticated_nexus_circuit_ref: CircuitRef,
    qa_project_name: str,
) -> None:
    """Test that we can run compile circuits for hypertket."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    compiled_circuits = qnx.compile(
        programs=[_authenticated_nexus_circuit_ref],
        name=f"qnexus_integration_test_compile_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
        hypertket_config=HyperTketConfig(),
    )

    assert len(compiled_circuits) == 1
    assert isinstance(compiled_circuits[0], CircuitRef)


def test_submit_execute(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that we can run an execute job in Nexus, wait for the job to complete and
    obtain the results from the execution."""

    config = qnx.AerConfig()

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)

    execute_job_ref = qnx.start_execute_job(
        programs=[my_circ],
        name=f"qnexus_integration_test_execute_job_{datetime.now()}",
        project=my_proj,
        backend_config=config,
        n_shots=[10],
    )

    assert isinstance(execute_job_ref, ExecuteJobRef)

    qnx.jobs.wait_for(execute_job_ref)

    execute_results = qnx.jobs.results(execute_job_ref)

    assert len(execute_results) == 1

    assert isinstance(execute_results[0].get_input(), CircuitRef)

    assert isinstance(execute_results[0].download_result(), BackendResult)

    assert isinstance(execute_results[0].download_backend_info(), BackendInfo)

    pj_ref = qnx.jobs.get(id=execute_job_ref.id)
    assert pj_ref.backend_config == config


def test_execute(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that we can run the utility execute function and get the results of the execution."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)

    backend_results = qnx.execute(
        programs=[my_circ],
        name=f"qnexus_integration_test_execute_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
        n_shots=[10],
    )

    assert len(backend_results) == 1
    assert isinstance(backend_results[0], BackendResult)
    assert isinstance(backend_results[0].get_counts(), Counter)


def test_get_results_for_incomplete_execute(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that we can run an execute job in Nexus, and fetch complete results for
    an otherwise errored/incomplete job."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)

    my_q_systems_circuit = qnx.circuits.upload(
        circuit=Circuit(2, 2).ZZPhase(0.5, 0, 1).measure_all(),
        name="qa_q_systems_circuit",
        project=my_proj,
    )

    execute_job_ref = qnx.start_execute_job(
        programs=[my_circ, my_q_systems_circuit],
        name=f"qnexus_integration_test_execute_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
        n_shots=[10, 10],
    )

    assert isinstance(execute_job_ref, ExecuteJobRef)

    with pytest.raises(qnx_exc.JobError):
        qnx.jobs.wait_for(execute_job_ref)

    incomplete_results = qnx.jobs.results(execute_job_ref, allow_incomplete=True)

    # wait for the ZZPhase circuit execution to complete
    for _ in range(10):
        incomplete_results = qnx.jobs.results(execute_job_ref, allow_incomplete=True)
        if len(incomplete_results) > 0:
            break
        sleep(10)

    # we expect the CX circuit to fail on H1-1LE, but the ZZPhase circuit should succeed
    assert len(incomplete_results) == 1

    assert isinstance(incomplete_results[0].get_input(), CircuitRef)
    assert isinstance(incomplete_results[0].download_result(), BackendResult)
    assert isinstance(incomplete_results[0].download_backend_info(), BackendInfo)


def test_wait_for_raises_on_job_error(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that if a job errors, the wait_for function raises an Exception."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    # Circuit not compiled for H1-1LE, so we expect it to error when executed
    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)

    failing_job_ref = qnx.start_execute_job(
        programs=[my_circ],
        name=f"qnexus_integration_test_failing_job_{datetime.now()}",
        project=my_proj,
        n_shots=[10],
        backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
    )

    with pytest.raises(qnx_exc.ResourceFetchFailed):
        qnx.jobs.results(failing_job_ref)

    with pytest.raises(qnx_exc.JobError):
        qnx.jobs.wait_for(failing_job_ref)


def test_results_not_available_error(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that we can't get the results of a job until it is complete."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)
    execute_job_ref = qnx.start_execute_job(
        programs=[my_circ],
        name=f"qnexus_integration_test_waiting_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
        n_shots=[10],
    )

    with pytest.raises(qnx_exc.ResourceFetchFailed):
        qnx.jobs.results(execute_job_ref)

    qnx.jobs.wait_for(execute_job_ref)

    execute_results = qnx.jobs.results(execute_job_ref)

    assert len(execute_results) == 1

    assert isinstance(execute_results[0].get_input(), CircuitRef)

    assert isinstance(execute_results[0].download_result(), BackendResult)

    assert isinstance(execute_results[0].download_backend_info(), BackendInfo)


def test_submit_under_user_group(
    _authenticated_nexus_circuit_ref: CircuitRef,
    qa_project_name: str,
    qa_circuit_name: str,
) -> None:
    """Test that a user can submit jobs under a user_group that
    they belong to.

    Requires that the test user is a member of a group called:
    'QA_IntegrationTestGroup',
    and not a member of a group called:
    'made_up_group'.
    """

    fake_group = "made_up_group"

    my_proj = qnx.projects.get(name_like=qa_project_name)

    with pytest.raises(qnx_exc.ResourceCreateFailed) as exc:
        qnx.start_compile_job(
            programs=[_authenticated_nexus_circuit_ref],
            name=f"qnexus_integration_test_compile_job_{datetime.now()}",
            project=my_proj,
            backend_config=qnx.AerConfig(),
            user_group=fake_group,
        )
        assert exc.value.message == f"Not a member of any group with name: {fake_group}"

    qnx.start_compile_job(
        programs=[_authenticated_nexus_circuit_ref],
        name=f"qnexus_integration_test_compile_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
        user_group="QA_IntegrationTestGroup",
    )

    my_circ = qnx.circuits.get(name_like=qa_circuit_name, project=my_proj)

    with pytest.raises(qnx_exc.ResourceCreateFailed):
        qnx.start_execute_job(
            programs=[my_circ],
            name=f"qnexus_integration_test_execute_job_{datetime.now()}",
            project=my_proj,
            backend_config=qnx.AerConfig(),
            n_shots=[10],
            user_group="made_up_group",
        )
        assert exc.value.message == f"Not a member of any group with name: {fake_group}"

    qnx.start_execute_job(
        programs=[my_circ],
        name=f"qnexus_integration_test_execute_job_{datetime.now()}",
        project=my_proj,
        backend_config=qnx.AerConfig(),
        n_shots=[10],
        user_group="QA_IntegrationTestGroup",
    )



"""Tests related to running jobs against QSys devices."""

from datetime import datetime
from typing import Any, cast

from guppylang import guppy  # type: ignore
from guppylang.std.builtins import result
from guppylang.std.quantum import cx, h, measure, qubit, x, z
from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from quantinuum_schemas.models.backend_config import SeleneQuestConfig

import qnexus as qnx
from qnexus.models.references import HUGRRef


def prepare_teleportation() -> Any:
    """Prepares the teleportation circuit."""

    @guppy
    def bell() -> tuple[qubit, qubit]:
        """Constructs a bell state."""
        q1, q2 = qubit(), qubit()
        h(q1)
        cx(q1, q2)
        return q1, q2

    @guppy
    def main() -> None:
        src = qubit()
        x(src)
        alice, bob = bell()

        cx(src, alice)
        h(src)
        if measure(alice):
            x(bob)
        if measure(src):
            z(bob)

        result("teleported", measure(bob))  # type: ignore

    return main.compile()  # type: ignore


def test_guppy_execution(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test the execution of a guppy program
    on a next-generation QSys device."""

    n_qubits = 3
    n_shots = 10

    project_ref = qnx.projects.get_or_create(name=qa_project_name)

    hugr_ref = qnx.hugr.upload(
        hugr_package=prepare_teleportation(),
        name="QA testing teleportation program",
        project=project_ref,
    )

    job_ref = qnx.start_execute_job(
        programs=[hugr_ref],
        n_shots=[n_shots],
        backend_config=SeleneQuestConfig(n_qubits=n_qubits),
        project=project_ref,
        name=f"QA Test QSys job from {datetime.now()}",
    )

    qnx.jobs.wait_for(job_ref)

    results = qnx.jobs.results(job_ref)

    assert len(results) == 1
    result_ref = results[0]

    assert isinstance(result_ref.download_backend_info(), BackendInfo)
    assert isinstance(result_ref.get_input(), HUGRRef)

    assert result_ref.get_input().id == hugr_ref.id

    qsys_result = cast(QsysResult, result_ref.download_result())
    assert len(qsys_result.results) == n_shots
    assert qsys_result.results[0].entries[0][0] == "teleported"
    assert qsys_result.results[0].entries[0][1] == 1

    # check some QsysResults functionality
    assert len(qsys_result.collated_counts().items()) > 0



"""Test that we can set properties on a project, and use them when creating/filtering resources."""

from datetime import datetime
from typing import OrderedDict

from pytket.circuit import Circuit

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict


def test_property_creation_and_filtering(
    _authenticated_nexus: None,
) -> None:
    """Test that we can create resources with property values
    and filter resources by their property values."""

    project_name = f"QA_test_project_properties_{datetime.now()}"

    my_new_project = qnx.projects.create(name=project_name)

    test_property_name_1 = "QA_test_prop_1"
    test_property_name_2 = "QA_test_prop_2"
    test_property_name_3 = "QA_test_prop_3"
    test_property_name_4 = "QA_test_prop_4"

    qnx.projects.add_property(
        name=test_property_name_1,
        property_type="bool",
        project=my_new_project,
        description="A test property for my QA test project",
    )

    qnx.projects.add_property(
        name=test_property_name_2,
        property_type="string",
        project=my_new_project,
        description="Another test property for my QA test project",
    )

    qnx.projects.add_property(
        name=test_property_name_3,
        property_type="int",
        project=my_new_project,
        description="Another test property for my QA test project",
    )

    qnx.projects.add_property(
        name=test_property_name_4,
        property_type="float",
        project=my_new_project,
        description="Another test property for my QA test project",
    )

    test_props = qnx.projects.get_properties(my_new_project)

    assert len(test_props) == 4

    circuit_name = f"QA_test_circuit_with_props_{datetime.now()}"

    my_circ = Circuit(2, 2).H(0).CX(0, 1).measure_all()

    qnx.circuits.upload(
        circuit=my_circ,
        name=circuit_name,
        project=my_new_project,
        properties=OrderedDict(
            {
                test_property_name_1: True,
                test_property_name_2: "test_string",
                test_property_name_3: 42,
                test_property_name_4: 3.15,
            }
        ),
    )

    qnx.circuits.upload(
        circuit=my_circ,
        name=circuit_name,
        project=my_new_project,
        properties=OrderedDict(
            {
                test_property_name_1: False,
            }
        ),
    )

    my_circuit_refs = qnx.circuits.get_all(
        project=my_new_project,
        properties=PropertiesDict(
            {
                test_property_name_1: True,
                test_property_name_2: "test_string",
                test_property_name_3: 42,
                test_property_name_4: 3.15,
            }
        ),
    )

    assert my_circuit_refs.count() == 1
    my_circuit_ref = my_circuit_refs.list()[0]

    assert my_circuit_ref.annotations.properties[test_property_name_1] is True
    assert my_circuit_ref.annotations.properties[test_property_name_2] == "test_string"

    qnx.projects.update(my_new_project, archive=True)
    qnx.projects.delete(my_new_project)



"""Test basic functionality relating to the hugr module."""

from datetime import datetime
from pathlib import Path

from hugr.package import Package

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import HUGRRef


def test_hugr_create_and_update(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can create a hugr and add a property value."""

    my_proj = qnx.projects.get(name_like=qa_project_name)

    hugr_name = f"QA_test_hugr_{datetime.now()}"

    hugr_path = Path("integration/data/hugr_example.hugr").resolve()
    hugr_package = Package.from_bytes(hugr_path.read_bytes())

    my_new_hugr = qnx.hugr.upload(
        hugr_package=hugr_package, name=hugr_name, project=my_proj
    )

    assert isinstance(my_new_hugr, HUGRRef)

    test_property_name = "QA_test_prop"
    test_prop_value = "foo"

    updated_hugr_ref = qnx.hugr.update(
        ref=my_new_hugr,
        properties=PropertiesDict({test_property_name: test_prop_value}),
    )

    assert (
        updated_hugr_ref.annotations.properties[test_property_name] == test_prop_value
    )


def test_hugr_download(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_hugr_name: str,
) -> None:
    """Test that valid HUGR can be extracted from an uploaded HUGR module."""

    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_hugr_ref = qnx.hugr.get(name_like=qa_hugr_name, project=my_proj)

    hugr_package = my_hugr_ref.download_hugr()
    assert isinstance(hugr_package, Package)


def test_hugr_get_by_id(
    _authenticated_nexus: None,
    qa_project_name: str,
    qa_hugr_name: str,
) -> None:
    """Test that we can get a HUGRRef by its ID."""
    my_proj = qnx.projects.get(name_like=qa_project_name)
    my_hugr_ref = qnx.hugr.get(name_like=qa_hugr_name, project=my_proj)

    hugr_ref_by_id = qnx.hugr.get(id=my_hugr_ref.id)

    assert hugr_ref_by_id == my_hugr_ref


def test_hugr_get_all(
    _authenticated_nexus: None,
    qa_project_name: str,
) -> None:
    """Test that we can get all HUGRRefs in a project."""
    my_proj = qnx.projects.get(name_like=qa_project_name)

    hugrs = qnx.hugr.get_all(project=my_proj)

    assert hugrs.count() >= 1
    assert isinstance(hugrs.list()[0], HUGRRef)



"""Test basic functionality relating to the auth module."""

from io import StringIO
from typing import Any

import pytest
from httpx import ConnectError

import qnexus as qnx
from qnexus.client import get_nexus_client
from qnexus.client.auth import login_no_interaction
from qnexus.client.utils import read_token
from qnexus.config import CONFIG
from qnexus.exceptions import AuthenticationError


def test_credential_login_full_flow(
    monkeypatch: Any,
) -> None:
    """Test that we can delete access tokens, login using credentials and
    delete tokens once again."""
    username = CONFIG.qa_user_email
    pwd = CONFIG.qa_user_password

    qnx.logout()
    with pytest.raises(FileNotFoundError):
        read_token(token_type="access_token")
        read_token(token_type="refresh_token")

    with pytest.raises(AuthenticationError):
        qnx.users.get_self()

    # fake user input from stdin
    monkeypatch.setattr("sys.stdin", StringIO(username + "\n"))
    monkeypatch.setattr("getpass.getpass", lambda prompt: pwd)

    qnx.login_with_credentials()

    assert read_token(token_type="access_token") != ""
    assert read_token(token_type="refresh_token") != ""

    qnx.users.get_self()

    qnx.logout()
    with pytest.raises(FileNotFoundError):
        read_token(token_type="access_token")
        read_token(token_type="refresh_token")

    with pytest.raises(AuthenticationError):
        qnx.users.get_self()

    # Login again to make sure credentials are in the system for the other tests
    # fake user input from stdin
    monkeypatch.setattr("sys.stdin", StringIO(username + "\n"))
    monkeypatch.setattr("getpass.getpass", lambda prompt: pwd)
    qnx.login_with_credentials()


@pytest.mark.skip(reason="Not implemented")
def test_device_code_flow_login_full_flow() -> None:
    """Test the flow for logging in with the browser."""


def test_domain_switch() -> None:
    """Set that we can reset the domain, login and not
    for tokens/URL to be dynamically loaded."""

    username = CONFIG.qa_user_email
    pwd = CONFIG.qa_user_password

    login_no_interaction(username, pwd)

    qnx.users.get_self()

    original_domain = CONFIG.domain

    # fake domain will reset the client value
    qnx.logout()
    fake_domain = "fake_nexus.com"
    CONFIG.domain = fake_domain
    assert fake_domain in str(get_nexus_client(reload=True).base_url)

    with pytest.raises(ConnectError):
        qnx.users.get_self()

    # setting it again will update the client without any restart required
    CONFIG.domain = original_domain
    login_no_interaction(username, pwd)
    assert original_domain in str(get_nexus_client().base_url)

    qnx.users.get_self()
